# coding: utf-8

from __future__ import absolute_import

import sys
import time
import unittest
from collections import namedtuple

import httpretty
import pytest
import reactivex as rx
from reactivex import operators as ops

import influxdb_client
from influxdb_client import WritePrecision, InfluxDBClient, VERSION
from influxdb_client.client.exceptions import InfluxDBError
from influxdb_client.client.write.point import Point
from influxdb_client.client.write_api import WriteOptions, WriteApi, PointSettings


class BatchingWriteTest(unittest.TestCase):

    def setUp(self) -> None:
        httpretty.enable()
        httpretty.reset()

        conf = influxdb_client.configuration.Configuration()
        conf.host = "http://localhost"
        conf.debug = False

        self.influxdb_client = InfluxDBClient(url=conf.host, token="my-token")

        self.write_options = WriteOptions(batch_size=2, flush_interval=5_000, retry_interval=3_000)
        self._write_client = WriteApi(influxdb_client=self.influxdb_client, write_options=self.write_options)

    def tearDown(self) -> None:
        self._write_client.close()
        httpretty.disable()

    def test_batch_size(self):
        httpretty.register_uri(httpretty.POST, uri="http://localhost/api/v2/write", status=204)
        httpretty.register_uri(httpretty.POST, uri="http://localhost/api/v2/write", status=204)

        self._write_client.write("my-bucket", "my-org",
                                 ["h2o_feet,location=coyote_creek level\\ water_level=1 1",
                                  "h2o_feet,location=coyote_creek level\\ water_level=2 2",
                                  "h2o_feet,location=coyote_creek level\\ water_level=3 3",
                                  "h2o_feet,location=coyote_creek level\\ water_level=4 4"])

        time.sleep(1)

        _requests = httpretty.httpretty.latest_requests

        self.assertEqual(2, len(_requests))
        _request1 = "h2o_feet,location=coyote_creek level\\ water_level=1 1\n" \
                    "h2o_feet,location=coyote_creek level\\ water_level=2 2"
        _request2 = "h2o_feet,location=coyote_creek level\\ water_level=3 3\n" \
                    "h2o_feet,location=coyote_creek level\\ water_level=4 4"

        self.assertEqual(_request1, _requests[0].parsed_body)
        self.assertEqual(_request2, _requests[1].parsed_body)
        pass

    def test_subscribe_wait(self):
        httpretty.register_uri(httpretty.POST, uri="http://localhost/api/v2/write", status=204)

        self._write_client.write("my-bucket", "my-org", "h2o_feet,location=coyote_creek level\\ water_level=1 1")
        self._write_client.write("my-bucket", "my-org", "h2o_feet,location=coyote_creek level\\ water_level=2 2")

        time.sleep(1)

        _requests = httpretty.httpretty.latest_requests

        self.assertEqual(1, len(_requests))

        _request = "h2o_feet,location=coyote_creek level\\ water_level=1 1\n" \
                   "h2o_feet,location=coyote_creek level\\ water_level=2 2"

        self.assertEqual(_request, _requests[0].parsed_body)

    def test_batch_size_group_by(self):
        httpretty.register_uri(httpretty.POST, uri="http://localhost/api/v2/write", status=204)

        self._write_client.write("my-bucket", "my-org",
                                 "h2o_feet,location=coyote_creek level\\ water_level=1 1")

        self._write_client.write("my-bucket", "my-org",
                                 "h2o_feet,location=coyote_creek level\\ water_level=2 2",
                                 write_precision=WritePrecision.S)

        self._write_client.write("my-bucket", "my-org-a",
                                 "h2o_feet,location=coyote_creek level\\ water_level=3 3")

        self._write_client.write("my-bucket", "my-org-a",
                                 "h2o_feet,location=coyote_creek level\\ water_level=4 4")

        self._write_client.write("my-bucket2", "my-org-a",
                                 "h2o_feet,location=coyote_creek level\\ water_level=5 5")

        self._write_client.write("my-bucket", "my-org-a",
                                 "h2o_feet,location=coyote_creek level\\ water_level=6 6")

        time.sleep(1)

        _requests = httpretty.httpretty.latest_requests

        self.assertEqual(5, len(_requests))

        self.assertEqual("h2o_feet,location=coyote_creek level\\ water_level=1 1", _requests[0].parsed_body)
        self.assertEqual("ns", _requests[0].querystring["precision"][0])
        self.assertEqual("my-bucket", _requests[0].querystring["bucket"][0])

        self.assertEqual("h2o_feet,location=coyote_creek level\\ water_level=2 2", _requests[1].parsed_body)
        self.assertEqual("s", _requests[1].querystring["precision"][0])
        self.assertEqual("my-bucket", _requests[1].querystring["bucket"][0])

        self.assertEqual("h2o_feet,location=coyote_creek level\\ water_level=3 3\n"
                         "h2o_feet,location=coyote_creek level\\ water_level=4 4", _requests[2].parsed_body)
        self.assertEqual("ns", _requests[2].querystring["precision"][0])
        self.assertEqual("my-bucket", _requests[2].querystring["bucket"][0])

        self.assertEqual("h2o_feet,location=coyote_creek level\\ water_level=5 5", _requests[3].parsed_body)
        self.assertEqual("ns", _requests[3].querystring["precision"][0])
        self.assertEqual("my-bucket2", _requests[3].querystring["bucket"][0])

        self.assertEqual("h2o_feet,location=coyote_creek level\\ water_level=6 6", _requests[4].parsed_body)
        self.assertEqual("ns", _requests[4].querystring["precision"][0])
        self.assertEqual("my-bucket", _requests[4].querystring["bucket"][0])

        pass

    def test_flush_interval(self):
        httpretty.register_uri(httpretty.POST, uri="http://localhost/api/v2/write", status=204)

        self._write_client.write("my-bucket", "my-org",
                                 ["h2o_feet,location=coyote_creek level\\ water_level=1 1",
                                  "h2o_feet,location=coyote_creek level\\ water_level=2 2"])

        time.sleep(1)
        self.assertEqual(1, len(httpretty.httpretty.latest_requests))

        self._write_client.write("my-bucket", "my-org", "h2o_feet,location=coyote_creek level\\ water_level=3 3")

        time.sleep(2)

        self.assertEqual(1, len(httpretty.httpretty.latest_requests))

        time.sleep(3)

        self.assertEqual(2, len(httpretty.httpretty.latest_requests))

        self.assertEqual("h2o_feet,location=coyote_creek level\\ water_level=3 3",
                         httpretty.httpretty.latest_requests[1].parsed_body)

    def test_jitter_interval(self):
        self._write_client.close()
        self._write_client = WriteApi(influxdb_client=self.influxdb_client,
                                      write_options=WriteOptions(batch_size=2, flush_interval=5_000,
                                                                 jitter_interval=3_000))

        httpretty.register_uri(httpretty.POST, uri="http://localhost/api/v2/write", status=204)

        self._write_client.write("my-bucket", "my-org",
                                 ["h2o_feet,location=coyote_creek level\\ water_level=1 1",
                                  "h2o_feet,location=coyote_creek level\\ water_level=2 2"])

        time.sleep(3)
        self.assertEqual(1, len(httpretty.httpretty.latest_requests))

        self._write_client.write("my-bucket", "my-org", "h2o_feet,location=coyote_creek level\\ water_level=3 3")

        time.sleep(2)

        self.assertEqual(1, len(httpretty.httpretty.latest_requests))

        time.sleep(6)

        self.assertEqual(2, len(httpretty.httpretty.latest_requests))

        self.assertEqual("h2o_feet,location=coyote_creek level\\ water_level=3 3",
                         httpretty.httpretty.latest_requests[1].parsed_body)

    def test_retry_interval(self):

        self._write_client.close()

        # Set retry interval to 1_500
        self.write_options = WriteOptions(batch_size=2, flush_interval=5_000, retry_interval=1_500)
        self._write_client = WriteApi(influxdb_client=self.influxdb_client, write_options=self.write_options)

        httpretty.register_uri(httpretty.POST, uri="http://localhost/api/v2/write", status=204)
        httpretty.register_uri(httpretty.POST, uri="http://localhost/api/v2/write", status=503)
        httpretty.register_uri(httpretty.POST, uri="http://localhost/api/v2/write", status=429,
                               adding_headers={'Retry-After': '3'})
        httpretty.register_uri(httpretty.POST, uri="http://localhost/api/v2/write", status=503)

        self._write_client.write("my-bucket", "my-org",
                                 ["h2o_feet,location=coyote_creek level\\ water_level=1 1",
                                  "h2o_feet,location=coyote_creek level\\ water_level=2 2"])

        time.sleep(1)
        self.assertEqual(1, len(httpretty.httpretty.latest_requests), msg="first request immediately")

        time.sleep(3)
        self.assertEqual(2, len(httpretty.httpretty.latest_requests), msg="second request after delay_interval")

        time.sleep(3)
        self.assertEqual(3, len(httpretty.httpretty.latest_requests), msg="third request after Retry-After")

        time.sleep(37.5)
        self.assertEqual(4, len(httpretty.httpretty.latest_requests), msg="fourth after exponential delay = 1.5 * 5**2")

        self.assertEqual("h2o_feet,location=coyote_creek level\\ water_level=1 1\n"
                         "h2o_feet,location=coyote_creek level\\ water_level=2 2",
                         httpretty.httpretty.latest_requests[0].parsed_body)
        self.assertEqual("h2o_feet,location=coyote_creek level\\ water_level=1 1\n"
                         "h2o_feet,location=coyote_creek level\\ water_level=2 2",
                         httpretty.httpretty.latest_requests[1].parsed_body)
        self.assertEqual("h2o_feet,location=coyote_creek level\\ water_level=1 1\n"
                         "h2o_feet,location=coyote_creek level\\ water_level=2 2",
                         httpretty.httpretty.latest_requests[2].parsed_body)
        self.assertEqual("h2o_feet,location=coyote_creek level\\ water_level=1 1\n"
                         "h2o_feet,location=coyote_creek level\\ water_level=2 2",
                         httpretty.httpretty.latest_requests[3].parsed_body)

        pass

    def test_retry_interval_max_retries(self):
        httpretty.register_uri(httpretty.POST, uri="http://localhost/api/v2/write", status=429,
                               adding_headers={'Retry-After': '1'})

        self._write_client.close()
        self._write_client = WriteApi(influxdb_client=self.influxdb_client,
                                      write_options=WriteOptions(batch_size=2, flush_interval=5_000, max_retries=5))

        self._write_client.write("my-bucket", "my-org",
                                 ["h2o_feet,location=coyote_creek level\\ water_level=1 1",
                                  "h2o_feet,location=coyote_creek level\\ water_level=2 2"])

        time.sleep(8)

        self.assertEqual(6, len(httpretty.httpretty.latest_requests))

    def test_retry_disabled_max_retries(self):
        httpretty.register_uri(httpretty.POST, uri="http://localhost/api/v2/write", status=429,
                               adding_headers={'Retry-After': '1'})

        self._write_client.close()
        self._write_client = WriteApi(influxdb_client=self.influxdb_client,
                                      write_options=WriteOptions(max_retries=0, batch_size=2, flush_interval=1_000))

        self._write_client.write("my-bucket", "my-org",
                                 ["h2o_feet,location=coyote_creek level\\ water_level=1 1",
                                  "h2o_feet,location=coyote_creek level\\ water_level=2 2"])

        time.sleep(2)

        self.assertEqual(1, len(httpretty.httpretty.latest_requests))

    def test_retry_disabled_max_retry_time(self):
        httpretty.register_uri(httpretty.POST, uri="http://localhost/api/v2/write", status=429,
                               adding_headers={'Retry-After': '1'})

        self._write_client.close()
        self._write_client = WriteApi(influxdb_client=self.influxdb_client,
                                      write_options=WriteOptions(max_retry_time=0, batch_size=2, flush_interval=1_000))

        self._write_client.write("my-bucket", "my-org",
                                 ["h2o_feet,location=coyote_creek level\\ water_level=1 1",
                                  "h2o_feet,location=coyote_creek level\\ water_level=2 2"])

        time.sleep(5)

        self.assertEqual(1, len(httpretty.httpretty.latest_requests))

    def test_recover_from_error(self):
        httpretty.register_uri(httpretty.POST, uri="http://localhost/api/v2/write", status=204)
        httpretty.register_uri(httpretty.POST, uri="http://localhost/api/v2/write", status=400)

        self._write_client.write("my-bucket", "my-org",
                                 ["h2o_feet,location=coyote_creek",
                                  "h2o_feet,location=coyote_creek"])

        self._write_client.write("my-bucket", "my-org",
                                 ["h2o_feet,location=coyote_creek level\\ water_level=1 1",
                                  "h2o_feet,location=coyote_creek level\\ water_level=2 2"])

        time.sleep(1)

        _requests = httpretty.httpretty.latest_requests

        self.assertEqual(2, len(_requests))

        pass

    def test_record_types(self):
        httpretty.register_uri(httpretty.POST, uri="http://localhost/api/v2/write", status=204)

        # Record item
        _record = "h2o_feet,location=coyote_creek level\\ water_level=1 1"
        self._write_client.write("my-bucket", "my-org", _record)

        # Point item
        _point = Point("h2o_feet").tag("location", "coyote_creek").field("level water_level", 2.0).time(2)
        self._write_client.write("my-bucket", "my-org", _point)

        # Record list
        self._write_client.write("my-bucket", "my-org",
                                 ["h2o_feet,location=coyote_creek level\\ water_level=3 3",
                                  "h2o_feet,location=coyote_creek level\\ water_level=4 4"])

        # Point list
        _point1 = Point("h2o_feet").tag("location", "coyote_creek").field("level water_level", 5.0).time(5)
        _point2 = Point("h2o_feet").tag("location", "coyote_creek").field("level water_level", 6.0).time(6)
        self._write_client.write("my-bucket", "my-org", [_point1, _point2])

        # Observable
        _recordObs = "h2o_feet,location=coyote_creek level\\ water_level=7 7"
        _pointObs = Point("h2o_feet").tag("location", "coyote_creek").field("level water_level", 8.0).time(8)

        self._write_client.write("my-bucket", "my-org", rx.of(_recordObs, _pointObs))

        _data = rx \
            .range(9, 13) \
            .pipe(ops.map(lambda i: "h2o_feet,location=coyote_creek level\\ water_level={0} {0}".format(i)))
        self._write_client.write("my-bucket", "my-org", _data)

        # Dictionary item
        _dict = {"measurement": "h2o_feet", "tags": {"location": "coyote_creek"},
                 "time": 13, "fields": {"level water_level": 13.0}}
        self._write_client.write("my-bucket", "my-org", _dict)

        # Dictionary list
        _dict1 = {"measurement": "h2o_feet", "tags": {"location": "coyote_creek"},
                  "time": 14, "fields": {"level water_level": 14.0}}
        _dict2 = {"measurement": "h2o_feet", "tags": {"location": "coyote_creek"},
                  "time": 15, "fields": {"level water_level": 15.0}}
        self._write_client.write("my-bucket", "my-org", [_dict1, _dict2])

        # Bytes item
        _bytes = "h2o_feet,location=coyote_creek level\\ water_level=16 16".encode("utf-8")
        self._write_client.write("my-bucket", "my-org", _bytes)

        # Bytes list
        _bytes1 = "h2o_feet,location=coyote_creek level\\ water_level=17 17".encode("utf-8")
        _bytes2 = "h2o_feet,location=coyote_creek level\\ water_level=18 18".encode("utf-8")
        self._write_client.write("my-bucket", "my-org", [_bytes1, _bytes2])

        # Tuple
        _bytes3 = "h2o_feet,location=coyote_creek level\\ water_level=19 19".encode("utf-8")
        _bytes4 = "h2o_feet,location=coyote_creek level\\ water_level=20 20".encode("utf-8")
        self._write_client.write("my-bucket", "my-org", (_bytes3, _bytes4,))

        time.sleep(1)

        _requests = httpretty.httpretty.latest_requests

        self.assertEqual(10, len(_requests))

        self.assertEqual("h2o_feet,location=coyote_creek level\\ water_level=1 1\n"
                         "h2o_feet,location=coyote_creek level\\ water_level=2 2", _requests[0].parsed_body)
        self.assertEqual("h2o_feet,location=coyote_creek level\\ water_level=3 3\n"
                         "h2o_feet,location=coyote_creek level\\ water_level=4 4", _requests[1].parsed_body)
        self.assertEqual("h2o_feet,location=coyote_creek level\\ water_level=5 5\n"
                         "h2o_feet,location=coyote_creek level\\ water_level=6 6", _requests[2].parsed_body)
        self.assertEqual("h2o_feet,location=coyote_creek level\\ water_level=7 7\n"
                         "h2o_feet,location=coyote_creek level\\ water_level=8 8", _requests[3].parsed_body)
        self.assertEqual("h2o_feet,location=coyote_creek level\\ water_level=9 9\n"
                         "h2o_feet,location=coyote_creek level\\ water_level=10 10", _requests[4].parsed_body)
        self.assertEqual("h2o_feet,location=coyote_creek level\\ water_level=11 11\n"
                         "h2o_feet,location=coyote_creek level\\ water_level=12 12", _requests[5].parsed_body)
        self.assertEqual("h2o_feet,location=coyote_creek level\\ water_level=13 13\n"
                         "h2o_feet,location=coyote_creek level\\ water_level=14 14", _requests[6].parsed_body)
        self.assertEqual("h2o_feet,location=coyote_creek level\\ water_level=15 15\n"
                         "h2o_feet,location=coyote_creek level\\ water_level=16 16", _requests[7].parsed_body)
        self.assertEqual("h2o_feet,location=coyote_creek level\\ water_level=17 17\n"
                         "h2o_feet,location=coyote_creek level\\ water_level=18 18", _requests[8].parsed_body)
        self.assertEqual("h2o_feet,location=coyote_creek level\\ water_level=19 19\n"
                         "h2o_feet,location=coyote_creek level\\ water_level=20 20", _requests[9].parsed_body)

        pass

    def test_write_point_different_precision(self):
        httpretty.register_uri(httpretty.POST, uri="http://localhost/api/v2/write", status=204)

        _point1 = Point("h2o_feet").tag("location", "coyote_creek").field("level water_level", 5.0) \
            .time(5, WritePrecision.S)
        _point2 = Point("h2o_feet").tag("location", "coyote_creek").field("level water_level", 6.0) \
            .time(6, WritePrecision.NS)

        self._write_client.write("my-bucket", "my-org", [_point1, _point2])

        time.sleep(1)

        _requests = httpretty.httpretty.latest_requests

        self.assertEqual(2, len(_requests))

        self.assertEqual("h2o_feet,location=coyote_creek level\\ water_level=5 5", _requests[0].parsed_body)
        self.assertEqual("s", _requests[0].querystring["precision"][0])
        self.assertEqual("h2o_feet,location=coyote_creek level\\ water_level=6 6", _requests[1].parsed_body)
        self.assertEqual("ns", _requests[1].querystring["precision"][0])

    def test_write_result(self):
        httpretty.register_uri(httpretty.POST, uri="http://localhost/api/v2/write", status=204)

        _record = "h2o_feet,location=coyote_creek level\\ water_level=1 1"
        _result = self._write_client.write("my-bucket", "my-org", _record)

        self.assertEqual(None, _result)

    def test_del(self):
        httpretty.register_uri(httpretty.POST, uri="http://localhost/api/v2/write", status=204)

        _record = "h2o_feet,location=coyote_creek level\\ water_level=1 1"
        _result = self._write_client.write("my-bucket", "my-org", _record)

        self._write_client.close()

        _requests = httpretty.httpretty.latest_requests

        self.assertEqual(1, len(_requests))
        self.assertEqual("h2o_feet,location=coyote_creek level\\ water_level=1 1", _requests[0].parsed_body)

    def test_default_tags(self):
        self._write_client.close()

        self.id_tag = "132-987-655"
        self.customer_tag = "California Miner"

        self._write_client = WriteApi(influxdb_client=self.influxdb_client,
                                      write_options=WriteOptions(batch_size=1),
                                      point_settings=PointSettings(**{"id": self.id_tag,
                                                                      "customer": self.customer_tag}))

        httpretty.register_uri(httpretty.POST, uri="http://localhost/api/v2/write", status=204)

        _point1 = {"measurement": "h2o_feet", "tags": {"location": "coyote_creek"},
                   "time": "2009-11-10T22:00:00Z", "fields": {"water_level": 1.0}}

        _point_list = [_point1]

        self._write_client.write("my-bucket", "my-org", _point_list)

        time.sleep(1)

        requests = httpretty.httpretty.latest_requests
        self.assertEqual(1, len(requests))

        request = str(requests[0].body)
        self.assertNotEqual(-1, request.find('customer=California\\\\ Miner'))
        self.assertNotEqual(-1, request.find('id=132-987-655'))

    def test_user_agent_header(self):
        httpretty.register_uri(httpretty.POST, uri="http://localhost/api/v2/write", status=204)

        self._write_client.write("my-bucket", "my-org",
                                 ["h2o_feet,location=coyote_creek level\\ water_level=1 1",
                                  "h2o_feet,location=coyote_creek level\\ water_level=2 2"])

        time.sleep(1)

        requests = httpretty.httpretty.latest_requests
        self.assertEqual(1, len(requests))
        self.assertEqual(f'influxdb-client-python/{VERSION}', requests[0].headers['User-Agent'])

    def test_to_low_flush_interval(self):

        self._write_client.close()
        self._write_client = WriteApi(influxdb_client=self.influxdb_client,
                                      write_options=WriteOptions(batch_size=8,
                                                                 flush_interval=1,
                                                                 jitter_interval=1000))

        httpretty.register_uri(httpretty.POST, uri="http://localhost/api/v2/write", status=204)

        for i in range(50):
            val_one = float(i)
            val_two = float(i) + 0.5
            point_one = Point("OneMillis").tag("sensor", "sensor1").field("PSI", val_one).time(time=i)
            point_two = Point("OneMillis").tag("sensor", "sensor2").field("PSI", val_two).time(time=i)

            self._write_client.write("my-bucket", "my-org", [point_one, point_two])
            time.sleep(0.1)

        self._write_client.close()

        _requests = httpretty.httpretty.latest_requests

        for _request in _requests:
            body = _request.parsed_body
            self.assertTrue(body, msg="Parsed body should be not empty " + str(_request))

        httpretty.reset()

    def test_batching_data_frame(self):
        from influxdb_client.extras import pd

        httpretty.register_uri(httpretty.POST, uri="http://localhost/api/v2/write", status=204)
        httpretty.register_uri(httpretty.POST, uri="http://localhost/api/v2/write", status=204)

        data_frame = pd.DataFrame(data=[["coyote_creek", 1.0], ["coyote_creek", 2.0],
                                        ["coyote_creek", 3.0], ["coyote_creek", 4.0]],
                                  index=[1, 2, 3, 4],
                                  columns=["location", "level water_level"])

        self._write_client.write("my-bucket", "my-org", record=data_frame,
                                 data_frame_measurement_name='h2o_feet',
                                 data_frame_tag_columns=['location'])

        time.sleep(1)

        _requests = httpretty.httpretty.latest_requests

        self.assertEqual(2, len(_requests))
        _request1 = "h2o_feet,location=coyote_creek level\\ water_level=1.0 1\n" \
                    "h2o_feet,location=coyote_creek level\\ water_level=2.0 2"
        _request2 = "h2o_feet,location=coyote_creek level\\ water_level=3.0 3\n" \
                    "h2o_feet,location=coyote_creek level\\ water_level=4.0 4"

        self.assertEqual(_request1, _requests[0].parsed_body)
        self.assertEqual(_request2, _requests[1].parsed_body)

    def test_named_tuple(self):
        self._write_client.close()
        self._write_client = WriteApi(influxdb_client=self.influxdb_client,
                                      write_options=WriteOptions(batch_size=1))

        httpretty.register_uri(httpretty.POST, uri="http://localhost/api/v2/write", status=204)

        Factory = namedtuple('Factory', ['measurement', 'position', 'customers'])
        factory = Factory(measurement='factory', position="central europe", customers=123456)

        self._write_client.write("my-bucket", "my-org", factory,
                                 record_measurement_key="measurement",
                                 record_tag_keys=["position"],
                                 record_field_keys=["customers"])

        time.sleep(1)

        _requests = httpretty.httpretty.latest_requests

        self.assertEqual(1, len(_requests))
        self.assertEqual("factory,position=central\\ europe customers=123456i", _requests[0].parsed_body)

    @pytest.mark.skipif(sys.version_info < (3, 8), reason="requires python3.8 or higher")
    def test_data_class(self):
        self._write_client.close()
        self._write_client = WriteApi(influxdb_client=self.influxdb_client,
                                      write_options=WriteOptions(batch_size=1))

        httpretty.register_uri(httpretty.POST, uri="http://localhost/api/v2/write", status=204)

        from dataclasses import dataclass

        @dataclass
        class Car:
            engine: str
            type: str
            speed: float

        car = Car('12V-BT', 'sport-cars', 125.25)
        self._write_client.write("my-bucket", "my-org",
                                 record=car,
                                 record_measurement_name="performance",
                                 record_tag_keys=["engine", "type"],
                                 record_field_keys=["speed"])

        time.sleep(1)

        _requests = httpretty.httpretty.latest_requests

        self.assertEqual(1, len(_requests))
        self.assertEqual("performance,engine=12V-BT,type=sport-cars speed=125.25", _requests[0].parsed_body)

    def test_success_callback(self):
        httpretty.register_uri(httpretty.POST, uri="http://localhost/api/v2/write", status=204)

        class SuccessCallback(object):
            def __init__(self):
                self.conf = None
                self.data = None

            def __call__(self, conf: (str, str, str), data: str):
                self.conf = conf
                self.data = data

        callback = SuccessCallback()

        self._write_client.close()
        self._write_client = WriteApi(influxdb_client=self.influxdb_client,
                                      write_options=WriteOptions(batch_size=2), success_callback=callback)

        self._write_client.write("my-bucket", "my-org",
                                 ["h2o_feet,location=coyote_creek water_level=1 1",
                                  "h2o_feet,location=coyote_creek water_level=2 2"])

        time.sleep(1)
        _requests = httpretty.httpretty.latest_requests
        self.assertEqual(1, len(_requests))
        self.assertEqual("h2o_feet,location=coyote_creek water_level=1 1\n"
                         "h2o_feet,location=coyote_creek water_level=2 2", _requests[0].parsed_body)

        self.assertEqual(b"h2o_feet,location=coyote_creek water_level=1 1\n"
                         b"h2o_feet,location=coyote_creek water_level=2 2", callback.data)
        self.assertEqual("my-bucket", callback.conf[0])
        self.assertEqual("my-org", callback.conf[1])
        self.assertEqual("ns", callback.conf[2])

    def test_error_callback(self):
        httpretty.register_uri(httpretty.POST, uri="http://localhost/api/v2/write", status=400)

        class ErrorCallback(object):
            def __init__(self):
                self.conf = None
                self.data = None
                self.error = None

            def __call__(self, conf: (str, str, str), data: str, error: InfluxDBError):
                self.conf = conf
                self.data = data
                self.error = error

        callback = ErrorCallback()

        self._write_client.close()
        self._write_client = WriteApi(influxdb_client=self.influxdb_client,
                                      write_options=WriteOptions(batch_size=2), error_callback=callback)

        self._write_client.write("my-bucket", "my-org",
                                 ["h2o_feet,location=coyote_creek water_level=1 x",
                                  "h2o_feet,location=coyote_creek water_level=2 2"])

        time.sleep(1)
        _requests = httpretty.httpretty.latest_requests
        self.assertEqual(1, len(_requests))
        self.assertEqual("h2o_feet,location=coyote_creek water_level=1 x\n"
                         "h2o_feet,location=coyote_creek water_level=2 2", _requests[0].parsed_body)

        self.assertEqual(b"h2o_feet,location=coyote_creek water_level=1 x\n"
                         b"h2o_feet,location=coyote_creek water_level=2 2", callback.data)
        self.assertEqual("my-bucket", callback.conf[0])
        self.assertEqual("my-org", callback.conf[1])
        self.assertEqual("ns", callback.conf[2])
        self.assertIsNotNone(callback.error)
        self.assertIsInstance(callback.error, InfluxDBError)
        self.assertEqual(400, callback.error.response.status)

    @pytest.mark.timeout(timeout=20)
    def test_error_callback_exception(self):
        httpretty.register_uri(httpretty.POST, uri="http://localhost/api/v2/write", status=400)

        class ErrorCallback(object):
            def __init__(self):
                self.conf = None
                self.data = None
                self.error = None

            def __call__(self, conf: (str, str, str), data: str, error: InfluxDBError):
                self.conf = conf
                self.data = data
                self.error = error
                raise Exception('Test generated an error')


        callback = ErrorCallback()

        self._write_client.close()
        self._write_client = WriteApi(influxdb_client=self.influxdb_client,
                                      write_options=WriteOptions(batch_size=2, max_close_wait=2_000), error_callback=callback)

        self._write_client.write("my-bucket", "my-org",
                                 ["h2o_feet,location=coyote_creek water_level=1 x",
                                  "h2o_feet,location=coyote_creek water_level=2 2"])

        time.sleep(1)
        _requests = httpretty.httpretty.latest_requests
        self.assertEqual(1, len(_requests))
        self.assertEqual("h2o_feet,location=coyote_creek water_level=1 x\n"
                         "h2o_feet,location=coyote_creek water_level=2 2", _requests[0].parsed_body)

        self.assertEqual(b"h2o_feet,location=coyote_creek water_level=1 x\n"
                         b"h2o_feet,location=coyote_creek water_level=2 2", callback.data)
        self.assertEqual("my-bucket", callback.conf[0])
        self.assertEqual("my-org", callback.conf[1])
        self.assertEqual("ns", callback.conf[2])
        self.assertIsNotNone(callback.error)
        self.assertIsInstance(callback.error, InfluxDBError)
        self.assertEqual(400, callback.error.response.status)

        self._write_client.close()


    def test_retry_callback(self):
        httpretty.register_uri(httpretty.POST, uri="http://localhost/api/v2/write", status=204)
        httpretty.register_uri(httpretty.POST, uri="http://localhost/api/v2/write", status=429, adding_headers={'Retry-After': '1'})
        httpretty.register_uri(httpretty.POST, uri="http://localhost/api/v2/write", status=503, adding_headers={'Retry-After': '1'})

        class RetryCallback(object):
            def __init__(self):
                self.count = 0
                self.conf = None
                self.data = None
                self.error = None

            def __call__(self, conf: (str, str, str), data: str, error: InfluxDBError):
                self.count += 1
                self.conf = conf
                self.data = data
                self.error = error

        callback = RetryCallback()

        self._write_client.close()
        self._write_client = WriteApi(influxdb_client=self.influxdb_client,
                                      write_options=WriteOptions(batch_size=2), retry_callback=callback)

        self._write_client.write("my-bucket", "my-org",
                                 ["h2o_feet,location=coyote_creek water_level=1 1",
                                  "h2o_feet,location=coyote_creek water_level=2 2"])

        time.sleep(3)
        _requests = httpretty.httpretty.latest_requests
        self.assertEqual(3, len(_requests))
        self.assertEqual("h2o_feet,location=coyote_creek water_level=1 1\n"
                         "h2o_feet,location=coyote_creek water_level=2 2", _requests[0].parsed_body)

        self.assertEqual(2, callback.count)
        self.assertEqual(b"h2o_feet,location=coyote_creek water_level=1 1\n"
                         b"h2o_feet,location=coyote_creek water_level=2 2", callback.data)
        self.assertEqual("my-bucket", callback.conf[0])
        self.assertEqual("my-org", callback.conf[1])
        self.assertEqual("ns", callback.conf[2])
        self.assertIsNotNone(callback.error)
        self.assertIsInstance(callback.error, InfluxDBError)
        self.assertEqual(429, callback.error.response.status)


if __name__ == '__main__':
    unittest.main()



"""Utils to get right Date parsing function."""
import datetime
import threading
from datetime import timezone as tz

from dateutil import parser

date_helper = None

lock_ = threading.Lock()


class DateHelper:
    """
    DateHelper to groups different implementations of date operations.

    If you would like to serialize the query results to custom timezone, you can use following code:

    .. code-block:: python

        from influxdb_client.client.util import date_utils
        from influxdb_client.client.util.date_utils import DateHelper
        import dateutil.parser
        from dateutil import tz

        def parse_date(date_string: str):
            return dateutil.parser.parse(date_string).astimezone(tz.gettz('ETC/GMT+2'))

        date_utils.date_helper = DateHelper()
        date_utils.date_helper.parse_date = parse_date
    """

    def __init__(self, timezone: datetime.tzinfo = tz.utc) -> None:
        """
        Initialize defaults.

        :param timezone: Default timezone used for serialization "datetime" without "tzinfo".
                         Default value is "UTC".
        """
        self.timezone = timezone

    def parse_date(self, date_string: str):
        """
        Parse string into Date or Timestamp.

        :return: Returns a :class:`datetime.datetime` object or compliant implementation
                 like :class:`class 'pandas._libs.tslibs.timestamps.Timestamp`
        """
        pass

    def to_nanoseconds(self, delta):
        """
        Get number of nanoseconds in timedelta.

        Solution comes from v1 client. Thx.
        https://github.com/influxdata/influxdb-python/pull/811
        """
        nanoseconds_in_days = delta.days * 86400 * 10 ** 9
        nanoseconds_in_seconds = delta.seconds * 10 ** 9
        nanoseconds_in_micros = delta.microseconds * 10 ** 3

        return nanoseconds_in_days + nanoseconds_in_seconds + nanoseconds_in_micros

    def to_utc(self, value: datetime):
        """
        Convert datetime to UTC timezone.

        :param value: datetime
        :return: datetime in UTC
        """
        if not value.tzinfo:
            return self.to_utc(value.replace(tzinfo=self.timezone))
        else:
            return value.astimezone(tz.utc)


def get_date_helper() -> DateHelper:
    """
    Return DateHelper with proper implementation.

    If there is a 'ciso8601' than use 'ciso8601.parse_datetime' else use 'dateutil.parse'.
    """
    global date_helper
    if date_helper is None:
        with lock_:
            # avoid duplicate initialization
            if date_helper is None:
                _date_helper = DateHelper()
                try:
                    import ciso8601
                    _date_helper.parse_date = ciso8601.parse_datetime
                except ModuleNotFoundError:
                    _date_helper.parse_date = parser.parse
                date_helper = _date_helper

    return date_helper



import datetime
import time

import pytest

from influxdb_client import PermissionResource, Permission, Task, InfluxDBClient
from influxdb_client.rest import ApiException
from tests.base_test import BaseTest

TASK_FLUX = 'from(bucket: "my-bucket") |> range(start: -2m) |> last()'


class TasksApiTest(BaseTest):

    def setUp(self) -> None:
        super(TasksApiTest, self).setUp()

        self.organization = self.find_my_org()
        self.authorization = self.add_tasks_authorization(self.organization)
        self.client.close()

        self.client = InfluxDBClient(self.host, self.authorization.token, debug=self.conf.debug)
        self.tasks_api = self.client.tasks_api()

        tasks = self.tasks_api.find_tasks()
        for task in tasks:
            if task.name.endswith("-IT"):
                self.tasks_api.delete_task(task.id)

    def add_tasks_authorization(self, organization):
        resource = PermissionResource(org=organization.name, type="tasks")

        create_task = Permission(resource=resource, action="read")
        delete_task = Permission(resource=resource, action="write")

        org_resource = PermissionResource(type="orgs")
        create_org = Permission(resource=org_resource, action="write")
        read_org = Permission(resource=org_resource, action="read")

        user_resource = PermissionResource(type="users")
        create_users = Permission(resource=user_resource, action="write")

        label_resource = PermissionResource(type="labels")
        create_labels = Permission(resource=label_resource, action="write")

        auth_resource = PermissionResource(type="authorizations")
        create_auth = Permission(resource=auth_resource, action="write")

        bucket = self.client.buckets_api().find_bucket_by_name("my-bucket")
        bucket_resource = PermissionResource(org_id=organization.id, id=bucket.id, type="buckets")
        read_bucket = Permission(resource=bucket_resource, action="read")
        write_bucket = Permission(resource=bucket_resource, action="write")

        return self.client.authorizations_api().create_authorization(org_id=organization.id,
                                                                     permissions=[create_task, delete_task, create_org,
                                                                                  read_org, create_users, create_labels,
                                                                                  create_auth, read_bucket,
                                                                                  write_bucket])

    def test_create_task(self):
        task_name = self.generate_name("it_task")

        flux = \
            '''option task = {{ 
                name: "{task_name}",
                every: 1h
            }}
            {flux}
            '''.format(task_name=task_name, flux=TASK_FLUX)

        task = Task(id=0, name=task_name, org_id=self.organization.id, flux=flux, status="active",
                    description="Task Description")

        task = self.tasks_api.create_task(task)

        print(task)

        self.assertIsNotNone(task)
        self.assertGreater(len(task.id), 1)

        self.assertEqual(task.name, task_name)
        self.assertEqual(task.org_id, self.organization.id)
        self.assertEqual(task.status, "active")
        self.assertEqual(task.every, "1h")
        self.assertEqual(task.cron, None)
        self.assertEqualIgnoringWhitespace(task.flux, flux)

        self.assertEqual(task.description, "Task Description")

    def test_create_task_with_offset(self):
        task_name = self.generate_name("it_task")

        flux = \
            '''option task = {{ 
                name: "{task_name}",
                every: 1h,
                offset: 30m
            }}
            {flux}
            '''.format(task_name=task_name, flux=TASK_FLUX)

        task = Task(id=0, name=task_name, org_id=self.organization.id, flux=flux, status="active",
                    description="Task Description")

        task = self.tasks_api.create_task(task)

        print(task)

        self.assertIsNotNone(task)
        self.assertEqual(task.offset, "30m")

    def test_create_task_every(self):
        task_name = self.generate_name("it_task")
        task = self.tasks_api.create_task_every(task_name, TASK_FLUX, "1h", self.organization)

        self.assertIsNotNone(task)
        self.assertGreater(len(task.id), 1)

        self.assertEqual(task.name, task_name)
        self.assertEqual(task.org_id, self.organization.id)
        self.assertEqual(task.status, "active")
        self.assertEqual(task.every, "1h")
        self.assertEqual(task.cron, None)
        self.assertTrue(task.flux.startswith(TASK_FLUX))

    def test_create_task_cron(self):
        task_name = self.generate_name("it task")
        task = self.tasks_api.create_task_cron(task_name, TASK_FLUX, "0 2 * * *", self.organization.id)

        self.assertIsNotNone(task)
        self.assertGreater(len(task.id), 1)

        self.assertEqual(task.name, task_name)
        self.assertEqual(task.org_id, self.organization.id)
        self.assertEqual(task.status, "active")
        self.assertEqual(task.every, None)
        self.assertEqual(task.cron, "0 2 * * *")
        # self.assertEqualIgnoringWhitespace(task.flux, flux)

        self.assertTrue(task.flux.startswith(TASK_FLUX))
        # self.assertEqual(task.links, "active")

        links = task.links
        self.assertIsNotNone(task.links)
        self.assertEqual(links.logs, "/api/v2/tasks/" + task.id + "/logs")
        self.assertEqual(links.members, "/api/v2/tasks/" + task.id + "/members")
        self.assertEqual(links.owners, "/api/v2/tasks/" + task.id + "/owners")
        self.assertEqual(links.runs, "/api/v2/tasks/" + task.id + "/runs")
        self.assertEqual(links._self, "/api/v2/tasks/" + task.id)

        # TODO missing get labels
        self.assertEqual(links.labels, "/api/v2/tasks/" + task.id + "/labels")

    def test_create_with_import(self):
        task_name = self.generate_name("it task")
        task_flux = 'import "http"\n\n' \
                    'from(bucket: "iot_center")\n' \
                    '    |> range(start: -30d)\n' \
                    '    |> filter(fn: (r) => r._measurement == "environment")\n' \
                    '    |> aggregateWindow(every: 1h, fn: mean)'
        task = self.tasks_api.create_task_cron(task_name, task_flux, "10 0 * * * *", self.organization.id)

        self.assertIsNotNone(task.id)
        self.assertEqual(task.name, task_name)
        self.assertEqual(task.org_id, self.organization.id)
        self.assertEqual(task.status, "active")
        self.assertEqual(task.cron, "10 0 * * * *")
        self.assertTrue(task.flux.startswith(task_flux))
        self.assertTrue(task.flux.splitlines()[-1].startswith('option task = '))

    def test_find_task_by_id(self):
        task_name = self.generate_name("it task")
        task = self.tasks_api.create_task_cron(task_name, TASK_FLUX, "0 2 * * *", self.organization.id)

        task_by_id = self.tasks_api.find_task_by_id(task.id)
        self.assertEqual(task, task_by_id)

    @pytest.mark.skip(reason="https://github.com/influxdata/influxdb/issues/11590")
    def test_find_task_by_user_id(self):
        task_user = self.users_api.create_user(self.generate_name("TaskUser"))
        self.tasks_api.create_task_cron(self.generate_name("it_task"), TASK_FLUX, "0 2 * * *",
                                        self.organization.id)
        tasks = self.tasks_api.find_tasks_by_user(task_user_id=task_user.id)
        print(tasks)
        self.assertEqual(len(tasks), 1)

    def test_find_tasks_iter(self):
        task_name = self.generate_name("it task")
        num_of_tasks = 10

        for _ in range(num_of_tasks):
            self.tasks_api.create_task_cron(task_name, TASK_FLUX, "0 2 * * *", self.organization.id)

        def count_unique_ids(tasks):
            return len(set(map(lambda task: task.id, tasks)))

        # get tasks in 3-4 batches
        tasks = self.tasks_api.find_tasks_iter(name= task_name, limit= num_of_tasks // 3)
        self.assertEqual(count_unique_ids(tasks), num_of_tasks)

        # get tasks in one equaly size batch
        tasks = self.tasks_api.find_tasks_iter(name= task_name, limit= num_of_tasks)
        self.assertEqual(count_unique_ids(tasks), num_of_tasks)

        # get tasks in one batch
        tasks = self.tasks_api.find_tasks_iter(name= task_name, limit= num_of_tasks + 1)
        self.assertEqual(count_unique_ids(tasks), num_of_tasks)

        # get no tasks
        tasks = self.tasks_api.find_tasks_iter(name= task_name + "blah")
        self.assertEqual(count_unique_ids(tasks), 0)

        # skip some tasks
        *_, split_task = self.tasks_api.find_tasks(name= task_name, limit= num_of_tasks // 3)
        tasks = self.tasks_api.find_tasks_iter(name= task_name, limit= 3, after= split_task.id)
        self.assertEqual(count_unique_ids(tasks), num_of_tasks - num_of_tasks // 3)

    def test_delete_task(self):
        task = self.tasks_api.create_task_cron(self.generate_name("it_task"), TASK_FLUX, "0 2 * * *",
                                               self.organization.id)
        self.assertIsNotNone(task)

        self.tasks_api.delete_task(task.id)
        with pytest.raises(ApiException) as e:
            assert self.tasks_api.find_task_by_id(task_id=task.id)
        assert "failed to find task" in e.value.body

    def test_update_task(self):
        task_name = self.generate_name("it task")
        cron_task = self.tasks_api.create_task_cron(task_name, TASK_FLUX, "0 2 * * *", self.organization.id)

        flux = '''
        {flux}
        
        option task = {{
            name: "{task_name}",
            every: 3m
        }}
        '''.format(task_name=task_name, flux=TASK_FLUX)

        cron_task.cron = None
        cron_task.every = "3m"
        cron_task.status = "inactive"
        cron_task.description = "Updated description"

        updated_task = self.tasks_api.update_task(cron_task)
        time.sleep(1)

        self.assertIsNotNone(updated_task)
        self.assertGreater(len(updated_task.id), 1)

        self.assertEqual(updated_task.name, task_name)
        self.assertEqual(updated_task.org_id, cron_task.org_id)
        self.assertEqual(updated_task.status, "inactive")
        self.assertEqual(updated_task.every, "3m")
        self.assertEqual(updated_task.cron, None)
        self.assertIsNotNone(updated_task.updated_at)
        now = datetime.datetime.now()
        now.astimezone()
        self.assertLess(updated_task.updated_at, now.astimezone(tz=datetime.timezone.utc))
        self.assertEqualIgnoringWhitespace(updated_task.flux, flux)

        self.assertEqual(updated_task.description, "Updated description")

    def test_member(self):
        task = self.tasks_api.create_task_cron(self.generate_name("it_task"), TASK_FLUX, "0 2 * * *",
                                               self.organization.id)
        members = self.tasks_api.get_members(task_id=task.id)
        self.assertEqual(len(members), 0)
        user = self.users_api.create_user(self.generate_name("Luke Health"))

        resource_member = self.tasks_api.add_member(member_id=user.id, task_id=task.id)
        self.assertIsNotNone(resource_member)
        self.assertEqual(resource_member.id, user.id)
        self.assertEqual(resource_member.name, user.name)
        self.assertEqual(resource_member.role, "member")

        members = self.tasks_api.get_members(task_id=task.id)
        resource_member = members[0]
        self.assertEqual(len(members), 1)
        self.assertEqual(resource_member.id, user.id)
        self.assertEqual(resource_member.name, user.name)
        self.assertEqual(resource_member.role, "member")

        self.tasks_api.delete_member(member_id=user.id, task_id=task.id)
        members = self.tasks_api.get_members(task_id=task.id)
        self.assertEqual(len(members), 0)

    @pytest.mark.skip(reason="https://github.com/influxdata/influxdb/issues/19234")
    def test_owner(self):
        task = self.tasks_api.create_task_cron(self.generate_name("it_task"), TASK_FLUX, "0 2 * * *",
                                               self.organization.id)
        owners = self.tasks_api.get_owners(task_id=task.id)
        self.assertEqual(len(owners), 1)

        user = self.users_api.create_user(self.generate_name("Luke Health"))
        resource_member = self.tasks_api.add_owner(owner_id=user.id, task_id=task.id)

        self.assertIsNotNone(resource_member)
        self.assertEqual(resource_member.id, user.id)
        self.assertEqual(resource_member.name, user.name)
        self.assertEqual(resource_member.role, "owner")

        owners = self.tasks_api.get_owners(task_id=task.id)
        self.assertEqual(len(owners), 2)
        resource_member = owners[1]
        self.assertEqual(resource_member.id, user.id)
        self.assertEqual(resource_member.name, user.name)
        self.assertEqual(resource_member.role, "owner")

        self.tasks_api.delete_owner(owner_id=user.id, task_id=task.id)
        owners = self.tasks_api.get_owners(task_id=task.id)
        self.assertEqual(len(owners), 1)

    def test_runs(self):
        task_name = self.generate_name("it task")
        task = self.tasks_api.create_task_every(task_name, TASK_FLUX, "1s", self.organization)
        time.sleep(5)

        runs = self.tasks_api.get_runs(task_id=task.id, limit=10)
        self.assertGreater(len(runs), 2)

        success_runs = list(filter(lambda x: x.status == "success", runs))
        run = success_runs[0]
        self.assertIsNotNone(run.id)
        self.assertEqual(run.task_id, task.id)
        self.assertEqual(run.status, "success")
        now = datetime.datetime.now()
        self.assertLess(run.scheduled_for, now.astimezone(tz=datetime.timezone.utc))
        self.assertLess(run.started_at, now.astimezone(tz=datetime.timezone.utc))
        self.assertLess(run.finished_at, now.astimezone(tz=datetime.timezone.utc))
        self.assertIsNone(run.requested_at)
        self.assertIsNotNone(run.links)

        self.assertEqual(run.links.retry, "/api/v2/tasks/" + task.id + "/runs/" + run.id + "/retry")
        self.assertEqual(run.links._self, "/api/v2/tasks/" + task.id + "/runs/" + run.id)
        self.assertEqual(run.links.task, "/api/v2/tasks/" + task.id)

    def test_runs_not_exist(self):
        with pytest.raises(ApiException) as e:
            assert self.tasks_api.get_runs("020f755c3c082000")
        assert "task not found" in e.value.body

    def test_run_task_manually(self):
        task = self.tasks_api.create_task_every(self.generate_name("it task"), TASK_FLUX, "1s", self.organization)

        run = self.tasks_api.run_manually(task_id=task.id)
        print(run)

        self.assertIsNotNone(run)
        self.assertTrue(run.status, "scheduled")

    def test_run_task_manually_not_exist(self):
        with pytest.raises(ApiException) as e:
            assert self.tasks_api.run_manually(task_id="020f755c3c082000")
        assert "failed to force run" in e.value.body

    def test_retry_run(self):
        task = self.tasks_api.create_task_every(self.generate_name("it task"), TASK_FLUX, "1s", self.organization)

        time.sleep(5)

        runs = self.tasks_api.get_runs(task.id)
        self.assertGreater(len(runs), 1)

        run = self.tasks_api.retry_run(task_id=runs[0].task_id, run_id=runs[0].id)
        self.assertIsNotNone(run)
        self.assertEqual(run.task_id, runs[0].task_id)

        self.assertEqual(run.status, "scheduled")
        self.assertEqual(run.task_id, task.id)

    def test_retry_run_not_exists(self):
        task = self.tasks_api.create_task_every(self.generate_name("it task"), TASK_FLUX, "5s", self.organization)
        with pytest.raises(ApiException) as e:
            assert self.tasks_api.retry_run(task_id=task.id, run_id="020f755c3c082000")
        assert "failed to retry run" in e.value.body

    def test_logs(self):
        task = self.tasks_api.create_task_every(self.generate_name("it task"), TASK_FLUX, "3s", self.organization)
        time.sleep(6)

        logs = self.tasks_api.get_logs(task_id=task.id)

        for log in logs:
            self.assertIsNotNone(log.time)
            self.assertIsNotNone(log.message)
            print(log)

        self.tasks_api.delete_task(task_id=task.id)

    def test_logs_not_exist(self):
        with pytest.raises(ApiException) as e:
            assert self.tasks_api.get_logs(task_id="020f755c3c082000")
        assert "failed to find task logs" in e.value.body

    def test_run_logs(self):
        task = self.tasks_api.create_task_every(self.generate_name("it task"), TASK_FLUX, "1s", self.organization)
        time.sleep(5)
        runs = self.tasks_api.get_runs(task_id=task.id)
        self.assertGreater(len(runs), 0)

        logs = self.tasks_api.get_run_logs(run_id=runs[-1].id, task_id=task.id)
        self.assertGreater(len(logs), 0)

        successes = list(filter(lambda log: log.message.endswith("Completed(success)"), logs))
        self.assertGreaterEqual(len(successes), 1)

    def test_runs_not_exists(self):
        task = self.tasks_api.create_task_every(self.generate_name("it task"), TASK_FLUX, "1s", self.organization)

        with pytest.raises(ApiException) as e:
            assert self.tasks_api.get_run_logs(task_id=task.id, run_id="020f755c3c082000")
        assert "failed to find task logs" in e.value.body

    def test_cancel_run_not_exist(self):
        task = self.tasks_api.create_task_every(self.generate_name("it task"), TASK_FLUX, "1s", self.organization)
        time.sleep(5)
        runs = self.tasks_api.get_runs(task.id)

        with pytest.raises(ApiException) as e:
            assert self.tasks_api.cancel_run(task_id=task.id, run_id=runs[-1].id)
        assert "failed to cancel run" in e.value.body
        assert "run not found" in e.value.body

    def test_cancel_task_not_exist(self):
        with pytest.raises(ApiException) as e:
            assert self.tasks_api.cancel_run("020f755c3c082000", "020f755c3c082000")
        assert "failed to cancel run" in e.value.body
        assert "task not found" in e.value.body

    def test_get_run(self):
        task = self.tasks_api.create_task_every(self.generate_name("it task"), TASK_FLUX, "1s", self.organization)
        time.sleep(5)
        run = self.tasks_api.get_runs(task_id=task.id)[0]
        self.assertIsNotNone(run)
        run_by_id = self.tasks_api.get_run(task.id, run.id)
        self.assertIsNotNone(run_by_id)
        self.assertEqual(run.id, run_by_id.id)

    def test_clone(self):
        task = self.tasks_api.create_task_every(self.generate_name("it_task"), TASK_FLUX, "1h", self.organization)
        label = self.labels_api.create_label(self.generate_name("it_task"), self.organization.id, {
            "color": "green",
            "location": "west"
        })
        self.tasks_api.add_label(label.id, task.id)
        cloned = self.tasks_api.clone_task(task)
        self.assertNotEqual(task.id, cloned.id)
        self.assertEqual(task.flux, cloned.flux)
        labels = self.tasks_api.get_labels(cloned.id).labels
        self.assertEqual(1, len(labels))
        self.assertEqual(label.id, labels[0].id)

    def test_clone_new(self):
        task = self.tasks_api._create_task(self.generate_name("it_task"), TASK_FLUX, "1h", None, self.organization.id)
        cloned = self.tasks_api.clone_task(task)
        self.assertNotEqual(task.id, cloned.id)
        self.assertEqual(task.flux, cloned.flux)
